local util = require("util")
if script.active_mods["gvv"] then require("__gvv__.gvv")() end
local next = next

-- some global variables to this mod that do not need to persist between saves
local reprocess_entities
local reprocess_entity_initial_count
local reprocess_entities_processed
local reprocess_start_tick
local reprocess_fixed_entities
local reprocess_count_per_tick
local reprocess_update_interval
local reprocess_enabled = false
local invalid_inventory

local function tile_prototype_collides_with_entity_prototype(tile_prototype, entity_prototype)
	if not (tile_prototype.collision_mask_with_flags["not-colliding-with-itself"] and entity_prototype.collision_mask_with_flags["not-colliding-with-itself"] and table.compare(tile_prototype.collision_mask, entity_prototype.collision_mask)) then
		for layer, _ in pairs(tile_prototype.collision_mask) do
			if entity_prototype.collision_mask[layer] then return true end
		end
	end
	return false
end

local function removeInvalidTiles(entity, layer, keep_artificial, destroy_tiles, second_entity)
	local colliding_tiles = entity.surface.find_tiles_filtered{area = entity.bounding_box, collision_mask = layer}
	local replacement_tiles = {}
	for _, tile in pairs(colliding_tiles) do
		if tile.hidden_tile and not tile_prototype_collides_with_entity_prototype(game.tile_prototypes[tile.hidden_tile], entity.prototype) then
			if keep_artificial then
				-- Keeping artificial tiles
				return false
			else
				table.insert(replacement_tiles, {position = tile.position, name = tile.hidden_tile})
			end
		else
			-- If there's even 1 tile that we're colliding with that we can't deal with, then the whole thing should be aborted
			return false
		end
	end
	if next(replacement_tiles) ~= nil then
		if second_entity then -- Basically, we want to check that none of the replacement tiles are going to collide with the second_entity, otherwise we still need to abort
			for _, rtile in pairs(replacement_tiles) do
				if tile_prototype_collides_with_entity_prototype(game.tile_prototypes[rtile.name], second_entity.prototype) then return false end
			end
		end
		if not destroy_tiles then
			for _, tile in pairs(colliding_tiles) do
				entity.surface.spill_item_stack(tile.position, game.tile_prototypes[tile.name].items_to_place_this[1], true, nil, false)
			end
		end
		entity.surface.set_tiles(replacement_tiles)
	end
	return true
end

local function process_entity(entity, surface)
	local fixed = false
	if global.worms[entity.name] then
		local found_influences = surface.find_entities_filtered{name = "worm-hole-" .. entity.name, position = entity.position}
		local influence
		for _, found_influence in pairs(found_influences) do
			if table.compare(found_influence.position, entity.position) then
				influence = found_influence
				break
			end
		end
		if not influence then
			influence = surface.create_entity{name = "worm-hole-" .. entity.name, position = entity.position, force = entity.force, create_build_effect_smoke = false}
			fixed = true
		end
		if not removeInvalidTiles(influence, global.layer, false, true, entity) then
			influence.destroy()
			entity.destroy{raise_destroy = true}
			fixed = true
		end
	elseif global.worm_holes[entity.name] then -- Cleanup check; looks for orphaned worm-holes and removes them (in case something has destroyed worms without raising the appropriate events)
		local worms = surface.find_entities_filtered{name = entity.name:gsub("worm%-hole%-", ""), position = entity.position}
		local found = false
		for _, worm in pairs(worms) do
			if table.compare(worm.position, entity.position) then found = true break end
		end
		if not found then
			entity.destroy()
			fixed = true
		end
	elseif global.spawners[entity.name] then
		local found_influences = surface.find_entities_filtered{name = "spawner-influence-" .. entity.name, position = entity.position}
		local influence
		for _, found_influence in pairs(found_influences) do
			if table.compare(found_influence.position, entity.position) then
				influence = found_influence
				break
			end
		end
		if not influence then
			influence = surface.create_entity{name = "spawner-influence-" .. entity.name, position = entity.position, force = entity.force, create_build_effect_smoke = false}
			fixed = true
		end
		if not removeInvalidTiles(influence, global.layer, false, true, entity) then
			influence.destroy()
			entity.destroy{raise_destroy = true}
			fixed = true
		end
	elseif global.spawner_influences[entity.name] then -- Cleanup check; looks for orphaned spawner-influences and removes them (in case something has destroyed spawners without raising the appropriate events)
		local spawners = surface.find_entities_filtered{name = entity.name:gsub("spawner%-influence%-", ""), position = entity.position}
		local found = false
		for _, spawner in pairs(spawners) do
			if table.compare(spawner.position, entity.position) then found = true break end
		end
		if not found then
			entity.destroy()
			fixed = true
		end
	elseif entity.type == "resource" then
		if not removeInvalidTiles(entity, "resource-layer") then
			entity.destroy{raise_destroy = true}
			fixed = true
		end
	elseif entity.type ~= "tree" and entity.name ~= "tree-roots" then
		if not removeInvalidTiles(entity, global.layer) then
			entity.mine{inventory = invalid_inventory, force = true}
			fixed = true
		end
	elseif entity.type == "tree" then
		-- Check to see if we already have tree-roots at this tree's position.  If not, then we'll need to create it
		local tree_position = entity.position
		local found_tree_roots = surface.find_entities_filtered{name = "tree-roots", position = tree_position}
		local tree_roots
		for _, found_tree_root in pairs(found_tree_roots) do
			if table.compare(tree_position, found_tree_root.position) then
				tree_roots = found_tree_root
				break
			end
		end
		if not tree_roots then
			tree_roots = surface.create_entity{name = "tree-roots", position = tree_position}
			fixed = true
		end
		-- Now check for any colliding tiles.  If there are any, then the tree-roots and associated tree (and stump) should be removed.
		if not removeInvalidTiles(tree_roots, global.layer, true) then
			local stumps_proto = entity.prototype.remains_when_mined
			entity.mine{inventory = invalid_inventory, force = true}
			for _, stump_proto in pairs(stumps_proto) do
				if stump_proto.type == "corpse" and stump_proto.name:find("stump") then
					local stumps = surface.find_entities_filtered{name = stump_proto.name, position = tree_position}
					for _, stump in pairs(stumps) do
						if table.compare(stump.position, tree_position) then stump.destroy() break end
					end
				end
			end
			tree_roots.destroy()
			fixed = true
		end
	elseif entity.name == "tree-roots" then -- Cleanup check; looks for orphaned tree-roots and removes them (in case something has destroyed trees without raising the appropriate events)
		local trees = surface.find_entities_filtered{type = "tree", position = entity.position}
		local found = false
		for _, tree in pairs(trees) do
			if table.compare(tree.position, entity.position) then found = true break end
		end
		if not found then
			entity.destroy()
			fixed = true
		end
	end
	return fixed
end

local function get_entities(surfaces, area)
	local return_entities = {}
	local return_count = 0
	local affect_trees = settings.startup["restrictionsonartificialtiles_affecttrees"].value
	local affect_resources = settings.startup["restrictionsonartificialtiles_affectresources"].value
	local affect_worms = settings.startup["restrictionsonartificialtiles_affectworms"].value
	local affect_spawners = settings.startup["restrictionsonartificialtiles_affectspawners"].value

	for _, surface in pairs(game.surfaces) do
		local entities = surface.find_entities_filtered{collision_mask = global.layer}
		for _, entity in pairs(entities) do
			return_count = return_count + 1
			return_entities[return_count] = entity
		end

		-- tree-roots
		if affect_trees then
			local tree_roots = surface.find_entities_filtered{name = "tree-roots", area = area}
			for _, tree_root in pairs(tree_roots) do
				return_count = return_count + 1
				return_entities[return_count] = tree_root
			end
		end

		-- Some of the rest of these don't have the mod's collision mask layer, so we need to get them separately, too
		--Resources
		if affect_resources then
			local resources = surface.find_entities_filtered{type = "resource", collision_mask = "resource-layer", area = area}
			for _, resource in pairs(resources) do
				return_count = return_count + 1
				return_entities[return_count] = resource
			end
		end

		-- Worms & worm-holes
		if affect_worms then
			for name, _ in pairs(global.worms) do
				local worms = surface.find_entities_filtered{name = name, area = area}
				for _, worm in pairs(worms) do
					return_count = return_count + 1
					return_entities[return_count] = worm
				end
			end
			for name, _ in pairs(global.worm_holes) do
				local worm_holes = surface.find_entities_filtered{name = name, area = area}
				for _, worm_hole in pairs(worm_holes) do
					return_count = return_count + 1
					return_entities[return_count] = worm_hole
				end
			end
		end

		-- Spawners & spawner-influences
		if affect_spawners then
			for name, _ in pairs(global.spawners) do
				local spawners = surface.find_entities_filtered{name = name, area = area}
				for _, spawner in pairs(spawners) do
					return_count = return_count + 1
					return_entities[return_count] = spawner
				end
			end
			for name, _ in pairs(global.spawner_influences) do
				local spawner_influences = surface.find_entities_filtered{name = name, area = area}
				for _, spawner_influence in pairs(spawner_influences) do
					return_count = return_count + 1
					return_entities[return_count] = spawner_influence
				end
			end
		end
	end

	return return_entities, return_count
end

local function process_entities(entity_names, surfaces, area)
	local entities = get_entities(surfaces or game.surfaces, area)
	invalid_inventory = game.create_inventory(0)
	for _, entity in pairs(entities) do
		if entity.valid and entity_names[entity.name] then
			process_entity(entity, entity.surface)
		end
	end
	invalid_inventory.destroy()
end

local function get_entity_prototype_names()
	local entity_names = {}
	local prototypes = game.get_filtered_entity_prototypes{{filter = "collision-mask", mask = global.layer, mask_mode = "collides"}}
	for name, _ in pairs(prototypes) do
		entity_names[name] = true
	end

	-- These do not have the appropriate collision mask, so we need to do them separately
	-- Resources
	if settings.startup["restrictionsonartificialtiles_affectresources"].value then
		local resources = game.get_filtered_entity_prototypes{{filter = "type", type = "resource"}, {filter = "collision-mask", mask = "resource-layer", mask_mode = "collides", mode = "and"}}
		for name, _ in pairs(resources) do
			entity_names[name] = true
		end
	end

	-- Worms
	if settings.startup["restrictionsonartificialtiles_affectworms"].value then
		for name, v in pairs(global.worms) do
			entity_names[name] = v
		end
	end

	-- Spawners
	if settings.startup["restrictionsonartificialtiles_affectspawners"].value then
		for name, v in pairs(global.spawners) do
			entity_names[name] = v
		end
	end

	return entity_names
end

local function InitAndConfigChangedCommon()
	global.layer = game.technology_prototypes.restrictionsonartificialtiles_layernameholder.effects[1].item

	global.worms = {}
	global.worm_fail_factor = {}
	global.worm_holes = {}
	for _, effect in pairs(game.technology_prototypes.restrictionsonartificialtiles_wormsholder.effects) do
		global.worms[effect.item] = true
		global.worm_fail_factor[effect.item] = tonumber(game.item_prototypes[effect.item].order)
		global.worm_holes["worm-hole-" .. effect.item] = true
	end

	global.spawners = {}
	global.spawner_fail_factor = {}
	global.spawner_influences = {}
	for _, effect in pairs(game.technology_prototypes.restrictionsonartificialtiles_spawnersholder.effects) do
		global.spawners[effect.item] = true
		global.spawner_fail_factor[effect.item] = tonumber(game.item_prototypes[effect.item].order)
		global.spawner_influences["spawner-influence-" .. effect.item] = true
	end

	settings.global["restrictionsonartificialtiles_reprocessentities"] = {value = false}
end

local function onInit()
	InitAndConfigChangedCommon()
	global.entity_list = get_entity_prototype_names()
	process_entities(global.entity_list)
end

local function onConfigurationChanged(event)
	if event.mod_changes["RestrictionsOnArtificialTiles"] and event.mod_changes["RestrictionsOnArtificialTiles"].old_version ~= nil then
		InitAndConfigChangedCommon()

		local old_entity_list = {}
		for k,v in pairs(global.entity_list) do
			old_entity_list[k] = v
		end
		global.entity_list = get_entity_prototype_names()
		local new_entities = {}
		for k, v in pairs(global.entity_list) do
			if not old_entity_list[k] then
				new_entities[k] = v
			end
		end
		if next(new_entities) ~= nil then
			process_entities(new_entities)
		end
	end
end

local function replaceTreePlacer(entity)
	entity.surface.create_entity{name = entity.name:gsub("tree%-placer%-", ""), position = entity.position, force = entity.force, fast_replace = true, player = entity.last_user, raise_built = true, create_build_effect_smoke = false}
	entity.destroy()
end

local function BuiltTree(tree)
	if not tree.prototype.collision_mask[global.layer] then return end
	local tree_roots = tree.surface.create_entity{name = "tree-roots", position = tree.position}
	if not removeInvalidTiles(tree_roots, global.layer) then
		invalid_inventory = game.create_inventory(0)
		tree.mine{inventory = invalid_inventory, force = true} -- I need to find a way to cancel or refund the item that places the tree in the case of player/robot building
		tree_roots.destroy()
		invalid_inventory.destroy()
	end
end

local function BuiltBiters(entity)
	local influence
	local entity_factor
	if global.worms[entity.name] then
		entity_factor = global.worm_fail_factor[entity.name]
		influence = entity.surface.create_entity{name = "worm-hole-" .. entity.name, position = entity.position, force = entity.force, create_build_effect_smoke = false}
	elseif global.spawners[entity.name] then
		entity_factor = global.spawner_fail_factor[entity.name]
		influence = entity.surface.create_entity{name = "spawner-influence-" .. entity.name, position = entity.position, force = entity.force, create_build_effect_smoke = false}
	end

	local influenced_tiles = influence.surface.find_tiles_filtered{area = influence.bounding_box, collision_mask = "ground-tile"} -- Don't want water tiles in this
	local fail_chance_factors = {0, 0, 0}
	for k, tile in pairs(influenced_tiles) do
		local tile_factor = tile.prototype.walking_speed_modifier or 1
		if tile_factor > 1 then
			fail_chance_factors[1] = fail_chance_factors[1] + (tile_factor - 1)
			fail_chance_factors[2] = fail_chance_factors[2] + ((tile_factor - 1) * entity_factor)
		end
		fail_chance_factors[3] = fail_chance_factors[3] + 1
	end
	local fail_chance = (fail_chance_factors[1] / fail_chance_factors[3]) + (fail_chance_factors[2] / fail_chance_factors[3])
	local math_random = math.random()
	if ((fail_chance ~= 0) and (math_random <= fail_chance)) or not removeInvalidTiles(influence, global.layer, false, true) then
		influence.destroy()
		entity.destroy{raise_destroy = true}
	end
end

local function removeBiterInfluence(entity)
	local influences
	if global.worms[entity.name] then
		influences = entity.surface.find_entities_filtered{name = "worm-hole-" .. entity.name, position = entity.position}
	elseif global.spawners[entity.name] then
		influences = entity.surface.find_entities_filtered{name = "spawner-influence-" .. entity.name, position = entity.position}
	else
		return
	end
	for _, influence in pairs(influences) do
		if table.compare(influence.position, entity.position) then influence.destroy() break end
	end
end

local function onBuiltEntity(event)
	local entity = event.created_entity
	if entity.name:find("tree%-placer%-") then
		replaceTreePlacer(entity)
	elseif entity.type == "tree" then
		BuiltTree(entity)
	elseif global.worms[entity.name] or global.spawners[entity.name] then
		BuiltBiters(entity)
	end
end

local function ScriptRaisedBuilt(event)
	local entity = event.entity
	if event.entity.type == "tree" then
		BuiltTree(event.entity)
	elseif global.worms[entity.name] or global.spawners[entity.name] then
		BuiltBiters(entity)
	elseif event.entity.name:find("$tree%-placer%-") then
		replaceTreePlacer(event.created_entity)
	end
end

local function TreeRemovedOrDied(tree)
	local tree_roots = tree.surface.find_entities_filtered{name = "tree-roots", position = tree.position}
	for _, tree_root in pairs(tree_roots) do
		if table.compare(tree.position, tree_root.position) then tree_root.destroy() break end
	end
end

local function TreeRootsDestroyed(tree_roots)
	local trees = tree_roots.surface.find_entities_filtered{type = "tree", position = tree_roots.position}
	for _, tree in pairs(trees) do
		if table.compare(tree.position, tree_roots.position) then tree.destroy() break end
	end
end

local function onMinedEntity(event)
	local entity = event.entity
	if entity.type == "tree" then
		TreeRemovedOrDied(event.entity)
	elseif global.worms[entity.name] or global.spawners[entity.name] then
		removeBiterInfluence(entity)
	end
end

local function onEntityDied(event)
	local entity = event.entity
	if entity.type == "tree" then
		TreeRemovedOrDied(entity)
	elseif global.worms[entity.name] or global.spawners[entity.name] then
		removeBiterInfluence(entity)
	end
end

local function ScriptRaisedDestroy(event)
	local entity = event.entity
	if entity.type == "tree" then
		TreeRemovedOrDied(entity)
	elseif global.worms[entity.name] or global.spawners[entity.name] then
		removeBiterInfluence(entity)
	elseif entity.name == "tree-roots" then
		TreeRootsDestroyed(entity)
	end
end

local function onTick(event)
	if reprocess_entities and next(reprocess_entities) ~= nil then
		local count = 0
		local i, entity = next(reprocess_entities)
		while i and count < reprocess_count_per_tick do
			if process_entity(entity, entity.surface) then
				reprocess_fixed_entities = reprocess_fixed_entities + 1
			end
			local index = i
			i, entity = next(reprocess_entities, i)
			reprocess_entities[index] = nil
			reprocess_entities_processed = reprocess_entities_processed + 1
			count = count + 1
		end
	else
		local message = "Restrictions On Artificial Tiles: Reprocess of all entities finished.  Made " .. reprocess_fixed_entities .. " fixes in "
		local total_time = math.floor((((game.tick - reprocess_start_tick) / 60) * 1000) + 0.5) / 1000
		if total_time > 120 then
			total_time = math.floor(((total_time / 60) * 100) + 0.5) / 100
			message = message .. total_time .. " minutes."
		else
			message = message .. total_time .. " seconds."
		end
		game.print(message)
		settings.global["restrictionsonartificialtiles_reprocessentities"] = {value = false}
	end
	if reprocess_update_interval > 0 and ((event.tick - reprocess_start_tick) % (reprocess_update_interval * 3600)) == 0 then
		game.print("Restrictions On Artificial Tiles: " .. reprocess_entities_processed .. " entities processed out of " .. reprocess_entity_initial_count .. ".")
	end
end

local function onRuntimeModSettingChanged(event)
	if event.setting_type == "runtime-global" and event.setting:find("restrictionsonartificialtiles") then
		if event.setting == "restrictionsonartificialtiles_reprocessentities" then
			if settings.global["restrictionsonartificialtiles_reprocessentities"].value and not reprocess_enabled then
				reprocess_entities, reprocess_entity_initial_count = get_entities(game.surfaces, nil)
				reprocess_fixed_entities = 0
				script.on_event(defines.events.on_tick, onTick)
				reprocess_start_tick = game.tick - 1 -- -1 avoids first tick printing of update to console
				reprocess_entities_processed = 0
				reprocess_count_per_tick = settings.global["restrictionsonartificialtiles_countpertick"].value
				reprocess_update_interval = settings.global["restrictionsonartificialtiles_updateinterval"].value
				local message = "Restrictions On Artificial Tiles: Reprocess of all entities initiated.  " .. reprocess_entity_initial_count .. " entities will be processed at a rate of " .. reprocess_count_per_tick .. " per tick, and update status messages "
				if reprocess_update_interval > 0 then
					message = message .. "will be printed every " .. reprocess_update_interval
					if reprocess_update_interval == 1 then
						message = message .. " minute."
					else
						message = message .. " minutes."
					end
				else
					message = message .. "are disabled."
				end
				game.print(message)
			else
				reprocess_entities = nil
				script.on_event(defines.events.on_tick, nil)
			end
			reprocess_enabled = settings.global["restrictionsonartificialtiles_reprocessentities"].value
		elseif event.setting == "restrictionsonartificialtiles_countpertick" then
			if reprocess_enabled and reprocess_count_per_tick ~= settings.global["restrictionsonartificialtiles_countpertick"].value then
				game.print("Restrictions On Artificial Tiles: Reprocess entity count per tick changed from " .. reprocess_count_per_tick .. " to " .. settings.global["restrictionsonartificialtiles_countpertick"].value .. ".")
				reprocess_count_per_tick = settings.global["restrictionsonartificialtiles_countpertick"].value
			end
		elseif event.setting == "restrictionsonartificialtiles_updateinterval" then
			if reprocess_enabled and reprocess_update_interval ~= settings.global["restrictionsonartificialtiles_updateinterval"].value then
				game.print("Restrictions On Artificial Tiles: Reprocess status update interval changed from " .. reprocess_update_interval .. " to " .. settings.global["restrictionsonartificialtiles_updateinterval"].value .. " minute(s).")
				reprocess_update_interval = settings.global["restrictionsonartificialtiles_updateinterval"].value
			end
		elseif event.setting == "restrictionsonartificialtiles_debug" then

			-- ### Begin debug test ###
			local surface = game.surfaces[1]
			game.print("Collision layer: " .. global.layer)
			log("Collision layer: " .. global.layer)
			local entities = surface.find_entities_filtered{collision_mask = global.layer}
			game.print("Found entities by collision_mask only: " .. #entities)
			log("Found entities by collision_mask only: " .. #entities)
			local found = false
			local trees = 0
			for _, entity in pairs(entities) do
				if entity.name == "tree-roots" then
					found = true
					log(entity.name .. " : " .. serpent.line(entity.position) .. " : " .. serpent.line(entity.prototype.collision_mask))
				elseif entity.type == "tree" then
					trees = trees + 1
				end
			end
			if #entities > 0 and not found then
				game.print("Found ".. trees .. " trees, but no \"tree-roots\"")
				log("Found ".. trees .. " trees, but no \"tree-roots\"")
			elseif #entities > 0 and found then
				game.print("Found \"tree-roots\"")
			end

			local tree_roots = surface.find_entities_filtered{name = "tree-roots", collision_mask = global.layer}
			game.print("Found entities by name and collision_mask: " .. #tree_roots)
			log("Found entities by name and collision_mask: " .. #tree_roots)
			found = false
			for _, tree_root in pairs(tree_roots) do
				if tree_root.name == "tree-roots" then
					found = true
					log(tree_root.name .. " : " .. serpent.line(tree_root.position) .. " : " .. serpent.line(tree_root.prototype.collision_mask))
				end
			end
			if #entities > 0 and not found then
				game.print("No \"tree-roots\"")
				log("No \"tree-roots\"")
			elseif #entities > 0 and found then
				game.print("Found \"tree-roots\"")
			end

			local tree_roots2 = surface.find_entities_filtered{name = "tree-roots"}
			game.print("Found entities by name only: " .. #tree_roots2)
			log("Found entities by name only: " .. #tree_roots2)
			found = false
			for _, tree_root2 in pairs(tree_roots2) do
				if tree_root2.name == "tree-roots" then
					found = true
					log(tree_root2.name .. " : " .. serpent.line(tree_root2.position) .. " : " .. serpent.line(tree_root2.prototype.collision_mask))
				end
			end
			if #entities > 0 and not found then
				game.print("No \"tree-roots\"")
				log("No \"tree-roots\"")
			elseif #entities > 0 and found then
				game.print("Found \"tree-roots\"")
			end
			-- ### End debug test ###

		end
	end
end

local function onBiterBaseBuilt(event)
	if global.worms[event.entity.name] or global.spawners[event.entity.name] then
		BuiltBiters(event.entity)
	end
end

local function onChunkGenerated(event)
	process_entities(global.entity_list, {event.surface}, event.area)
end

script.on_init(onInit)
script.on_configuration_changed(onConfigurationChanged)
script.on_event(defines.events.on_built_entity, onBuiltEntity, {{filter = "type", type = "tree"}, {filter = "type", type = "offshore-pump"}, {filter = "type", type = "turret"}, {filter = "type", type = "unit-spawner"}})
script.on_event(defines.events.on_robot_built_entity, onBuiltEntity, {{filter = "type", type = "tree"}, {filter = "type", type = "offshore-pump"}, {filter = "type", type = "turret"}, {filter = "type", type = "unit-spawner"}})
script.on_event(defines.events.script_raised_built, ScriptRaisedBuilt, {{filter = "type", type = "tree"}, {filter = "type", type = "offshore-pump"}, {filter = "type", type = "turret"}, {filter = "type", type = "unit-spawner"}})
script.on_event(defines.events.on_player_mined_entity, onMinedEntity, {{filter = "type", type = "tree"}, {filter = "type", type = "turret"}, {filter = "type", type = "unit-spawner"}})
script.on_event(defines.events.on_robot_mined_entity, onMinedEntity, {{filter = "type", type = "tree"}, {filter = "type", type = "turret"}, {filter = "type", type = "unit-spawner"}})
script.on_event(defines.events.on_entity_died, onEntityDied, {{filter = "type", type = "tree"}, {filter = "type", type = "turret"}, {filter = "type", type = "unit-spawner"}})
script.on_event(defines.events.script_raised_destroy, ScriptRaisedDestroy, {{filter = "type", type = "tree"}, {filter = "name", name = "tree-roots"}, {filter = "type", type = "turret"}, {filter = "type", type = "unit-spawner"}})
script.on_event(defines.events.on_runtime_mod_setting_changed, onRuntimeModSettingChanged)
script.on_event(defines.events.on_biter_base_built, onBiterBaseBuilt)
script.on_event(defines.events.on_chunk_generated, onChunkGenerated)