local util = require("util")
local collision_mask_util = require("collision-mask-util")
require("__base__/prototypes/entity/pipecovers")

-- This mod's collision mask layer
local artificial_tile_collision_layer = collision_mask_util.get_first_unused_layer()

-- settings, so we don't have to make calls to settings.startup for every entity
local affect_trees = settings.startup["restrictionsonartificialtiles_affecttrees"].value
local affect_miners = settings.startup["restrictionsonartificialtiles_affectminers"].value
local affect_resources = settings.startup["restrictionsonartificialtiles_affectresources"].value
local affect_worms = settings.startup["restrictionsonartificialtiles_affectworms"].value
local affect_spawners = settings.startup["restrictionsonartificialtiles_affectspawners"].value
local affect_cliffs = settings.startup["restrictionsonartificialtiles_affectcliffs"].value

local data_entities = {}
-- dummy_item and dummy_tech used to convey the created collision mask layer of this mod to the control stage so it can be used to search for and find entities affected by this mod
local dummy_item = {
	type = "item",
	name = artificial_tile_collision_layer,
	icon = "__core__/graphics/icons/unknown.png",
	icon_size = 64,
	stack_size = 1,
	flags = { "hidden" }
}
table.insert(data_entities, dummy_item)

local dummy_tech = {
	type = "technology",
	name = "restrictionsonartificialtiles_layernameholder",
	icon = "__core__/graphics/icons/unknown.png",
	icon_size = 64,
	enabled = false,
	hidden = true,
	ignore_tech_cost_multiplier = true,
	effects = {
		{type = "give-item", item = dummy_item.name}
	},
	unit = {
		count = 1,
		ingredients = {
			{"automation-science-pack", 1},
			{"logistic-science-pack", 1},
			{"military-science-pack", 1},
			{"chemical-science-pack", 1},
			{"production-science-pack", 1},
			{"utility-science-pack", 1},
			{"space-science-pack", 1}
		},
		time = 1
	}
}
table.insert(data_entities, dummy_tech)

if affect_trees then
	-- Used to allow a larger collision radius for tiles around trees without affecting anything else
	local tree_roots = {
		type = "simple-entity",
		name = "tree-roots",
		icon = "__core__/graphics/icons/unknown.png",
		icon_size = 64,
		flags = {"placeable-neutral", "placeable-off-grid", "hidden"},
		collision_box = {{-0.95, -0.95}, {0.95, 0.95}},
		collision_mask = {"colliding-with-tiles-only", artificial_tile_collision_layer},
		picture = {
			layers = {
				{
					filename = "__core__/graphics/empty.png",
					size = 1
				}
			}
		}
	}
	table.insert(data_entities, tree_roots)
end

if affect_worms then
	-- Used to convey registered worms to the control stage as the collision mask layer of this mod is not added to them
	local dummy_worm_tech = {
		type = "technology",
		name = "restrictionsonartificialtiles_wormsholder",
		icon = "__core__/graphics/icons/unknown.png",
		icon_size = 64,
		enabled = false,
		hidden = true,
		ignore_tech_cost_multiplier = true,
		effects = {},
		unit = {
			count = 1,
			ingredients = {
				{"automation-science-pack", 1},
				{"logistic-science-pack", 1},
				{"military-science-pack", 1},
				{"chemical-science-pack", 1},
				{"production-science-pack", 1},
				{"utility-science-pack", 1},
				{"space-science-pack", 1}
			},
			time = 1
		}
	}
	table.insert(data_entities, dummy_worm_tech)
end

if affect_spawners then
	-- Used to convey registered spawners to the control stage as the collision mask layer of this mod is not added to them
	local dummy_spawner_tech = {
		type = "technology",
		name = "restrictionsonartificialtiles_spawnersholder",
		icon = "__core__/graphics/icons/unknown.png",
		icon_size = 64,
		enabled = false,
		hidden = true,
		ignore_tech_cost_multiplier = true,
		effects = {},
		unit = {
			count = 1,
			ingredients = {
				{"automation-science-pack", 1},
				{"logistic-science-pack", 1},
				{"military-science-pack", 1},
				{"chemical-science-pack", 1},
				{"production-science-pack", 1},
				{"utility-science-pack", 1},
				{"space-science-pack", 1}
			},
			time = 1
		}
	}
	table.insert(data_entities, dummy_spawner_tech)
end

data:extend(data_entities)

-- The actual function that adds the collision mask layer from this mod
local function add_collision_layer(prototype, layer)
	local collisions = collision_mask_util.get_mask(prototype)
	collision_mask_util.add_layer(collisions, layer)
	prototype.collision_mask = collisions
end

-- Global table of functions for modders to call to register their prototypes with this mod
RestrictionsOnArtificialTiles = {}

-- Register function for tiles
function RestrictionsOnArtificialTiles.Register_tile(tile)
	add_collision_layer(tile, artificial_tile_collision_layer)
	if affect_resources then
		add_collision_layer(tile, "resource-layer")
	end
	tile.check_collision_with_entities = true
	tile.decorative_removal_probability = 1
end

-- Register function for trees
function RestrictionsOnArtificialTiles.Register_tree(tree)
	if affect_trees then
		add_collision_layer(tree, artificial_tile_collision_layer)
		if not tree.exclude_from_placer then
			-- Used to ensure tree placement by the player (when enabled by another mod or via /editor) respects the radius the "tree-roots" entity will put around the tree once placed
			local tree_placer = {
				type = "offshore-pump",
				name = "tree-placer-" .. tree.name,
				localised_name = (tree.localised_name or tree.name),
				icon = tree.icon,
				icon_size = tree.icon_size,
				icon_mipmaps = tree.icon_mipmaps,
				flags = {"hide-alt-info"},
				collision_box = table.deepcopy(tree.collision_box),
				collision_mask = table.deepcopy(tree.collision_mask),
				center_collision_mask = {"water-tile", artificial_tile_collision_layer},
				fluid_box_tile_collision_test = {},
				adjacent_tile_collision_test = {},
				adjacent_tile_collision_mask = {artificial_tile_collision_layer},
				adjacent_tile_collision_box = {{-0.95, -0.95}, {0.95, 0.95}},
				fluid = "water",
				fluid_box = {
					pipe_connections = {
						{
							position = {0, tree.collision_box[2][2] + 0.01}
						}
					}
				},
				pumping_speed = 1,
			}

			if tree.flags then
				for i, flag in pairs(tree.flags) do
					tree_placer.flags[i + 1] = tree.flags[i]
				end
				if not util.list_to_map(tree.flags)["hidden"] then
					table.insert(tree.flags, "hidden")
				end
			end

			if tree.variations then
				tree_placer.picture = {layers = {table.deepcopy(tree.variations[1].trunk)}}
				tree_placer.picture.layers[2] = table.deepcopy(tree.variations[1].leaves)
				tree_placer.picture.layers[1].frame_count = 1
				tree_placer.picture.layers[2].frame_count = 1
				if tree_placer.picture.layers[1].hr_version then
					tree_placer.picture.layers[1].hr_version.frame_count = 1
					tree_placer.picture.layers[2].hr_version.frame_count = 1
				end
			elseif tree.pictures then
				tree_placer.picture = {layers = {table.deepcopy(tree.pictures[1])}}
			elseif tree.picture then
				tree_placer.picture = table.deepcopy(tree.picture)
			else
				tree_placer.picture = {
					layers = {
						{
							filename = "__base__/graphics/entity/wooden-chest/wooden-chest.png", -- TODO: gotta find something better.  But hey, at least it'll be obvious when it don't work, right? :P
							priority = "extra-high",
							width = 32,
							height = 36,
							shift = util.by_pixel(0.5, -2)
						}
					}
				}
			end

			if tree.integration_patch then
				tree_placer.integration_patch = table.deepcopy(tree.integration_patch)
			end

			if tree.placeable_by then
				tree_placer.placeable_by = table.deepcopy(tree.placeable_by)
				if #tree.placeable_by > 0 then
					for _, placeable in pairs(tree.placeable_by) do
						data.raw.item[placeable.item].place_result = tree_placer.name
					end
				else
					data.raw.item[tree.placeable_by.item].place_result = tree_placer.name
				end
			elseif data.raw.item[tree.name] then
				data.raw.item[tree.name].place_result = tree_placer.name
			elseif data.raw["item-with-entity-data"][tree.name] then
				data.raw["item-with-entity-data"][tree.name].place_result = tree_placer.name
			else
				for _, item in pairs(data.raw.item) do
					if item.place_result == tree.name then
						item.place_result = tree_placer.name
					end
				end
			end

			if tree.icons then
				tree_placer.icons = table.deepcopy(tree.icons)
			end

			data:extend({tree_placer})
		end
	end
end

-- Register function for mining-drills
function RestrictionsOnArtificialTiles.Register_miner(mining_drill)
	if affect_miners then
		add_collision_layer(mining_drill, artificial_tile_collision_layer)
	end
end

-- Register function for worms
function RestrictionsOnArtificialTiles.Register_worm(worm, fail_factor)
	if affect_worms then
		-- Collision boxes could be off-center, so this handles that.
		-- Big and behemoth have the same collision_box size despite different sprite sizes, so that's why the different factor.
		local x_increase = math.max(0.5, (worm.collision_box[2][1] - worm.collision_box[1][1]) * (worm.name:find("behemoth") and 0.5 or 0.25))
		local y_increase = math.max(0.5, (worm.collision_box[2][2] - worm.collision_box[1][2]) * (worm.name:find("behemoth") and 0.5 or 0.25))
		-- Used to allow a larger collision radius for tiles around worms without affecting anything else
		local worm_hole = {
			type = "simple-entity",
			name = "worm-hole-" .. worm.name,
			icon = "__core__/graphics/icons/unknown.png",
			icon_size = 64,
			flags = {"placeable-enemy", "placeable-off-grid", "hidden"},
			collision_box = {{worm.collision_box[1][1] - x_increase, worm.collision_box[1][2] - y_increase}, {worm.collision_box[2][1] + x_increase, worm.collision_box[2][2] + y_increase}},
			collision_mask = {"colliding-with-tiles-only", artificial_tile_collision_layer},
			picture = {
				layers = {
					{
						filename = "__core__/graphics/empty.png",
						size = 1
					}
				}
			}
		}

		-- Used to convey registered worms to the control stage as the collision mask layer of this mod is not added to them
		local dummy_worm_item = {
			type = "item",
			name = worm.name,
			icon = "__core__/graphics/icons/unknown.png",
			icon_size = 64,
			stack_size = 1,
			flags = { "hidden" },
			order = fail_factor
		}
		table.insert(data.raw.technology.restrictionsonartificialtiles_wormsholder.effects, {type = "give-item", item = dummy_worm_item.name})

		data:extend({worm_hole, dummy_worm_item})
	end
end

-- Register function for spawners
function RestrictionsOnArtificialTiles.Register_spawner(spawner, fail_factor)
	if affect_spawners then
		-- Collision boxes could be off-center, so this handles that.
		local x_increase = math.max(0.5, (spawner.collision_box[2][1] - spawner.collision_box[1][1]) * 0.125)
		local y_increase = math.max(0.5, (spawner.collision_box[2][2] - spawner.collision_box[1][2]) * 0.125)
		-- Used to allow a larger collision radius for tiles around spawners without affecting anything else
		local spawner_influence = {
			type = "simple-entity",
			name = "spawner-influence-" .. spawner.name,
			icon = "__core__/graphics/icons/unknown.png",
			icon_size = 64,
			flags = {"placeable-enemy", "placeable-off-grid", "hidden"},
			collision_box = {{spawner.collision_box[1][1] - x_increase, spawner.collision_box[1][2] - y_increase}, {spawner.collision_box[2][1] + x_increase, spawner.collision_box[2][2] + y_increase}},
			collision_mask = {"colliding-with-tiles-only", artificial_tile_collision_layer},
			picture = {
				layers = {
					{
						filename = "__core__/graphics/empty.png",
						size = 1
					}
				}
			}
		}

		-- Used to convey registered spawners to the control stage as the collision mask layer of this mod is not added to them
		local dummy_spawner_item = {
			type = "item",
			name = spawner.name,
			icon = "__core__/graphics/icons/unknown.png",
			icon_size = 64,
			stack_size = 1,
			flags = { "hidden" },
			order = fail_factor
		}
		table.insert(data.raw.technology.restrictionsonartificialtiles_spawnersholder.effects, {type = "give-item", item = dummy_spawner_item.name})

		data:extend({spawner_influence, dummy_spawner_item})
	end
end

-- Register function for cliffs
function RestrictionsOnArtificialTiles.Register_cliff(cliff)
	if affect_cliffs then
		add_collision_layer(cliff, artificial_tile_collision_layer)
	end
end

-- Register function for whatever else modders want registered.  Optional parameter "setting" for registering the prototype only if that related setting is enabled.
function RestrictionsOnArtificialTiles.Register_other(prototype, setting)
	if not setting or
		(setting == "trees" and affect_trees) or
		(setting == "miners" and affect_miners) or
		(setting == "resources" and affect_resources) or
		(setting == "worms" and affect_worms) or
		(setting == "spawners" and affect_spawners) or
		(setting == "cliffs" and affect_cliffs) then
		add_collision_layer(prototype, artificial_tile_collision_layer)
	end
end